import re
from typing import Callable, Any

def interpolate_docstring(names : dict[str, Any] = dict()) -> Callable[[Callable], Callable] :
    def inner(func : Callable) -> Callable:
        docstring : str | None = func.__doc__
        if docstring is not None:
            docstring = docstring.replace('{', '{{').replace('}', '}}')
            docstring = re.sub(r'\${{([^{}]*)}}', r'{\1}', docstring)
            g = { **func.__globals__, **names }
            func.__doc__ = eval('rf"""' + docstring + '"""', g)
        return func
    return inner




"""

Module containing utilities to store and manipulate
abstract filtered simplicial complexes.
		
"""
from __future__ import annotations
__all__ = ['FilteredComplex', 'Simplex', 'clique_complex', 'standard_simplex']
class FilteredComplex:
    """
    
    Class representing a filtered simplicial complex.
    """
    def __init__(self, n: int, k: int) -> None:
        """
        Construct a discrete filtered simplicial complex with default filtration time of 0.
        
        Args:
        	n : Number of vertices. Cannot be changed after initialisation.
        	k : Maximum dimension of a simplex that the complex can have. This parameter is required for memory efficiency, and cannot be changed after initialisation.
        """
    def __repr__(self) -> str:
        ...
    def add_simplex(self, vertices: list[int], filt_value: float) -> bool:
        """
        Add a simplex to a filtered simplicial complex.
        
        Args:
        	vertices : List of vertex labels corresponding to existing vertices in the complex.
        	filt_value : Filtration value to associate to the new simplex.
        
        Note:
        	Faces of the added simplex that are already present in the simplicial complex will have their filtration values reduced if necessary.
        """
    def get_label_from_vertex_labels(self, vertices: list[int]) -> int:
        """
        				Returns the dictionary key of a simplex with respect to its vertex labels sorted in ascending order, counting all
        				possible sorted subsequences of :math:`(0, ..., N-1)` of length :math:`k`, where :math:`N` is the number of vertices in the complex. The simplex need not be present in the simplicial complex.
        
        				Args:
        					vertices : List of vertex labels of the simplex.
        """
    def has_simplex(self, vertices: list[int]) -> bool:
        """
        				Check for membership of a simplex in the complex.
        
        				Args:
        					vertices : Vertex labels of the simplex to check for.
        """
    def is_filtration(self) -> bool:
        """
        			 Returns true if the filtration property is satisfied; that is, if each simplex has a filtration value at least as large as each of its faces.
        """
    def propagate_colours(self) -> None:
        """
        				Function to make sure that simplex colours are consistent with the colours of their vertices. You should call this whenever you change the colour of any vertices.
        """
    def propagate_filt_values(self, start_dim: int, upwards: bool = True) -> None:
        """
        Propagate filtration values upwards or downwards to ensure that every simplex appears after its faces. For example, setting the filtration values in dimension 1 and propagating upwards is akin to the Rips filtration.
        
        Args:
        	start_dim : Dimension from which to start propagating (exclusive).
        	upwards : If true then values are propagated upwards, downwards otherwise. Defaults to true.
        """
    def serialised(self) -> list[tuple[list[int], int, float, int]]:
        """
        				Serialised representation of the simplicial complex in a format suitable for persistent homology computations.
        
        				:return:
        					A list `x` of simplices in the simplicial complex ordered by dimension followed by filtration time. Each simplex :math:`\sigma` is represented by a tuple containing the following items.
        
        					1.  A list containing the indices in `x` of the facets of :math:`\sigma`, sorted in ascending order.
        					2.  The lexicographic key of :math:`\sigma` in the simplicial complex.
        					3.  The filtration time of :math:`\sigma`.
        					4.  The colour bitmask of :math:`\sigma`.
        """
    @property
    def dimension(self) -> int:
        """
        Current maximum dimension of a maximal simplex in the complex.
        """
    @property
    def max_dimension(self) -> int:
        """
        Maximum dimension of simplex that this complex can store. \
        Set during initialisation.
        """
    @property
    def max_filtration_time(self) -> float:
        """
        Current maximum dimension of a maximal simplex in the complex.
        """
    @property
    def num_simplices(self) -> int:
        """
        The total number of simplices in the complex.
        """
    @property
    def num_vertices(self) -> int:
        """
        Number of vertices in the simplicial complex. \
        Set during initialisation.
        """
    @property
    def simplices(self) -> list[dict[int, Simplex]]:
        """
        				A list such that ``simplices[k]`` is a dictionary of handles to the :math:`k`-simplices in the complex. The key of a :math:`k`-simplex in ``simplices[k]`` is the lexicographic index of that simplex with respect to its vertex labels sorted in ascending order, counting all possible sorted subsequences of :math:`(0, ..., N-1)` of length :math:`k`.
        """
class Simplex:
    """
    
    			Class representing a simplex in a filtered simplicial complex.
    		
    """
    def __repr__(self) -> str:
        ...
    def set_colour(self, colour: int) -> None:
        """
        				Change the colour of a vertex.
        
        				Raises:
        					ValueError: If the simplex is not a vertex or if `colour >=` :attr:`MaxColoursChromatic <chalc.chromatic.MaxColoursChromatic>`.
        
        				Tip:
        					It is recommended to call the member function :meth:`propagate_colours() <chalc.filtration.FilteredComplex.propagate_colours>` from the parent simplicial complex after changing the colour of a vertex.
        """
    @property
    def colours(self) -> int:
        """
        				Bitmask of colours in the simplex, where the rightmost bit represents the smallest colour index. More precisely, :math:`\mathrm{bitmask} = \sum_{c \in \mathrm{colours}(\sigma)} 2^c`. For example, a simplex having vertices of colours 0 and 1 has a colour bitmask of 3.
        """
    @property
    def dimension(self) -> int:
        """
        			Dimension of the simplex.
        """
    @property
    def facets(self) -> list[Simplex]:
        """
        				Read-only list of handles to the facets of the simplex.
        """
    @property
    def filtration_value(self) -> float:
        """
        				Filtration value of the simplex. If you modify this value, you should call :meth:`propagate_filt_values() <chalc.filtration.FilteredComplex.propagate_filt_values>` from the parent complex to ensure that filtration times remain monotonic.
        """
    @filtration_value.setter
    def filtration_value(self, arg0: float) -> None:
        ...
    @property
    def label(self) -> int:
        """
        				Label of the simplex in its parent filtered complex. A :math:`k`-simplex :math:`\sigma` is labelled by the lexicographic index of :math:`\sigma` with respect to its vertex labels sorted in ascending order, counting all possible sorted subsequences of :math:`(0, ..., N-1)` of length :math:`k`.
        """
    @property
    def vertices(self) -> list[int]:
        """
        				List of (sorted, ascending) vertex labels of the simplex.
        """
def clique_complex(n: int, k: int) -> FilteredComplex:
    """
    			Returns the :math:`k`-skeleton of the complete simplicial complex on :math:`n` vertices, with filtration values initialised to zero and all vertices coloured with the colour 0.
    """
def standard_simplex(n: int) -> FilteredComplex:
    """
    			Returns the filtered simplicial complex corresponding to the standard abstract :math:`n`-simplex, with filtration values initialised to zero and all vertices coloured with the colour 0.
    """



"""

Module containing geometry routines to compute chromatic Delaunay filtrations.

Attributes:
	MaxColoursChromatic (int): Maximum number of colours that can be handled by the methods in this module.
        
"""
from __future__ import annotations
import chalc.filtration
import numpy
import pybind11_stubgen.typing_ext
import typing
__all__ = ['MaxColoursChromatic', 'alpha', 'delaunay', 'delcech', 'delrips']
@typing.overload
def alpha(x: typing.Annotated[numpy.ndarray, numpy.float64, pybind11_stubgen.typing_ext.DynamicSize('m', 'n')], colours: typing.Annotated[numpy.ndarray, numpy.int32, pybind11_stubgen.typing_ext.DynamicSize('m', 1)]) -> tuple[chalc.filtration.FilteredComplex, bool]:
    ...
@typing.overload
def alpha(x: typing.Annotated[numpy.ndarray, numpy.float64, pybind11_stubgen.typing_ext.DynamicSize('m', 'n')], colours: list[int]) -> tuple[chalc.filtration.FilteredComplex, bool]:
    """
    Computes the chromatic alpha filtration of a coloured point cloud.
    
    Args:
    	x : Numpy matrix whose columns are points in the point cloud.
    	colours : List or numpy array of integers describing the colours of the points.
    
    Returns:
    	The chromatic alpha filtration and a boolean flag to indicate if numerical issues were encountered. In case of numerical issues, a warning is also raised.
    
    Raises:
    	ValueError: If any value in ``colours`` is >= :attr:`MaxColoursChromatic <chalc.chromatic.MaxColoursChromatic>` or < 0, or if the length of ``colours`` does not match the number of points.
    
    Notes:
    	This function is included for pedantic reasons. For most purposes you should instead consider using :func:`chalc.chromatic.delcech`, which is faster to compute, more numerically stable, and has the same persistent homology.
    
    See Also:
    	:func:`delrips`, :func:`delcech`
    """
@typing.overload
def delaunay(x: typing.Annotated[numpy.ndarray, numpy.float64, pybind11_stubgen.typing_ext.DynamicSize('m', 'n')], colours: typing.Annotated[numpy.ndarray, numpy.int32, pybind11_stubgen.typing_ext.DynamicSize('m', 1)]) -> chalc.filtration.FilteredComplex:
    ...
@typing.overload
def delaunay(x: typing.Annotated[numpy.ndarray, numpy.float64, pybind11_stubgen.typing_ext.DynamicSize('m', 'n')], colours: list[int]) -> chalc.filtration.FilteredComplex:
    """
    Returns the chromatic Delaunay triangulation of a coloured point cloud in Euclidean space.
    
    Args:
    	x : Numpy matrix whose columns are points in the point cloud.
    	colours : List or numpy array of integers describing the colours of the points.
    
    Raises:
    	ValueError: If any value in ``colours`` is >= :attr:`MaxColoursChromatic <chalc.chromatic.MaxColoursChromatic>` or < 0, or if the length of ``colours`` does not match the number of points.
    
    Returns:
    	The Delaunay triangulation.
    """
@typing.overload
def delcech(x: typing.Annotated[numpy.ndarray, numpy.float64, pybind11_stubgen.typing_ext.DynamicSize('m', 'n')], colours: typing.Annotated[numpy.ndarray, numpy.int32, pybind11_stubgen.typing_ext.DynamicSize('m', 1)]) -> tuple[chalc.filtration.FilteredComplex, bool]:
    ...
@typing.overload
def delcech(x: typing.Annotated[numpy.ndarray, numpy.float64, pybind11_stubgen.typing_ext.DynamicSize('m', 'n')], colours: list[int]) -> tuple[chalc.filtration.FilteredComplex, bool]:
    """
    Returns the chromatic Delaunay--Čech filtration of a coloured point cloud.
    
    Args:
    	x : Numpy matrix whose columns are points in the point cloud.
    	colours : List or numpy array of integers describing the colours of the points.
    
    Returns:
    	The chromatic Delaunay--Čech filtration and a boolean flag to indicate if numerical issues were encountered. In case of numerical issues, a warning is also raised.
    
    Raises:
    	ValueError : If any value in ``colours`` is >= :attr:`MaxColoursChromatic <chalc.chromatic.MaxColoursChromatic>` or < 0, or if the length of ``colours`` does not match the number of points.
    
    Notes:
    	The chromatic Delaunay--Čech filtration of the point cloud has the same set of simplices as the chromatic alpha filtration, but with Čech filtration times.
    
    See Also:
    	:func:`alpha`, :func:`delrips`
    """
@typing.overload
def delrips(x: typing.Annotated[numpy.ndarray, numpy.float64, pybind11_stubgen.typing_ext.DynamicSize('m', 'n')], colours: typing.Annotated[numpy.ndarray, numpy.int32, pybind11_stubgen.typing_ext.DynamicSize('m', 1)]) -> tuple[chalc.filtration.FilteredComplex, bool]:
    ...
@typing.overload
def delrips(x: typing.Annotated[numpy.ndarray, numpy.float64, pybind11_stubgen.typing_ext.DynamicSize('m', 'n')], colours: list[int]) -> tuple[chalc.filtration.FilteredComplex, bool]:
    """
    Computes the chromatic Delaunay--Rips filtration of a coloured point cloud.
    
    Args:
    	x : Numpy matrix whose columns are points in the point cloud.
    	colours : List or numpy array of integers describing the colours of the points.
    
    Returns:
    	The chromatic Delaunay--Rips filtration and a boolean flag to indicate if numerical issues were encountered. In case of numerical issues, a warning is also raised.
    
    Raises:
    	ValueError: If any value in ``colours`` is >= :attr:`MaxColoursChromatic <chalc.chromatic.MaxColoursChromatic>` or < 0, or if the length of ``colours`` does not match the number of points.
    
    Notes:
    	The chromatic Delaunay--Rips filtration of the point cloud has the same set of simplices as the chromatic alpha filtration, but with Vietoris--Rips filtration times. The convention used is that the filtration time of a simplex is half the maximum edge length in that simplex. With this convention, the chromatic Delaunay--Rips filtration and chromatic alpha filtration have the same persistence diagrams in degree zero.
    
    See Also:
    	:func:`alpha`, :func:`delcech`
    """
MaxColoursChromatic: int = 64



import sys, os
from pathlib import Path
from platform import python_version
from packaging.version import parse

# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here.
docs_path = Path(__file__).parent
project_root_dir = docs_path.parent
stubs_path = str((docs_path / 'stubs' ).resolve())
exts_path = str((docs_path / 'exts').resolve())
sys.path.insert(0, stubs_path)
sys.path.append(exts_path)

copyright = '2023, Abhinav Natarajan'
author = 'Abhinav Natarajan'

# Get the name and release from pyproject.toml
if parse(python_version()) >= parse('3.11'):
    from tomllib import loads as tomlread
else:
    from toml import loads as tomlread
with open(project_root_dir / "pyproject.toml") as f:
    proj_props = tomlread(f.read())['project']

project = proj_props['name']
release = '.'.join(map(str, parse(proj_props['version']).release))

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

maximum_signature_line_length = 160
toc_object_entries_show_parents = 'hide'
add_module_names = False
master_doc = "index"
templates_path = ['templates']
exclude_patterns = [
    '_build', 
    'templates', 
    'exts', 
    'static',
    # 'example.ipynb'
]

extensions = [
    'sphinx.ext.autosectionlabel',
    'sphinx.ext.napoleon',
    'sphinx.ext.autodoc',
    'sphinx.ext.autosummary',
    'autosummary_customise',
    'sphinx.ext.intersphinx',
    'nbsphinx',
    'sphinx_design',
    'sphinx_toolbox.sidebar_links',
    'sphinx_toolbox.github',
    'sphinx.ext.githubpages']

# autosectionlabel options
autosectionlabel_prefix_document = True

# napoleon options
napolean_include_init_with_doc = True
napoleon_attr_annotations = True

# autodoc options
autodoc_default_options = {
    'special-members': '',
    'undoc-members': False,
    'private-members': False,
    'imported-members': False,
}
autodoc_typehints = 'description'
autodoc_typehints_description_target = "documented_params"

#autosummary options
# autosummary_imported_members = True
# autosummary_ignore_module_all = False

# intersphinx options
intersphinx_mapping = {
    'python' : ('https://docs.python.org/3/', None),
    'numpy' : ('https://numpy.org/doc/stable', None),
    'matplotlib' : ('https://matplotlib.org/stable', None),
    'h5py' : ('https://docs.h5py.org/en/stable/', None)
}

# nbsphinx options
nbsphinx_execute = 'always'
os.environ['PYDEVD_DISABLE_FILE_VALIDATION'] = '1'

# toolbox.github options
github_username = 'abhinavnatarajan'
github_repository = 'chalc'

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_static_path = ['static']
html_css_files = [
    'css/custom.css'
]

html_theme = 'furo'
html_theme_options = {
    "source_repository": "https://github.com/abhinavnatarajan/chalc",
    "source_branch": "master",
    "source_directory": "docs/",
    "top_of_page_button": "edit"
}



import pybind11_stubgen
from pathlib import Path

if __name__ == '__main__':

    pybind11_stubgen.main()

    stubs_path = (Path(__file__).parents[0].resolve() / 'stubs').resolve()
    stubs = list(stubs_path.rglob('*.pyi'))
    for x in stubs:
        x.replace(x.with_suffix('.py'))



from __future__ import annotations

from collections.abc import Collection, Sequence
from itertools import product
from typing import Annotated

import matplotlib.animation as animation
import matplotlib.pyplot as plt
import numpy as np
import seaborn as sns
from matplotlib.axes import Axes
from matplotlib.figure import Figure
from pandas import DataFrame

from chalc.filtration import FilteredComplex

from ._utils import interpolate_docstring
from .sixpack import (
	DiagramEnsemble,
	_bitmask_to_colours,
	_colours_are_subset,
	_colours_to_bitmask,
	_num_colours_in_bitmask,
)

plt.rcParams["animation.html"] = "jshtml"

__doc__ = "Plotting and visualisation utilities."


def plot_sixpack(
	dgms: DiagramEnsemble,
	truncation: float | None = None,
	max_diagram_dimension: int | None = None,
	tolerance: float = 0,
) -> tuple[Figure, Annotated[np.ndarray, "Axes"]]:
	"""
	Plots the 6-pack of persistence diagrams returned by :func:`compute <.sixpack.compute>`.

	Args:
		dgms                  : The 6-pack of persistence diagrams.
		truncation            : The maximum entrance time upto which features are plotted. A sensible default will be calculated if not provided.
		max_diagram_dimension : The maximum homological dimension for which to plot features. If not provided, all dimensions will be included in the plots.
		tolerance             : Only features with persistence greater than this value will be plotted.
	"""
	if max_diagram_dimension is None:
		max_diagram_dimension = max(dgms.dimensions)
	dim_shift = {name: (1 if name == "ker" else 0) for name in DiagramEnsemble.diagram_names}
	max_dim = {
		name: (max_diagram_dimension if name != "rel" else max_diagram_dimension + 1)
		for name in DiagramEnsemble.diagram_names
	}
	plot_pos = {
		name: (val[1], val[0])
		for name, val in zip(DiagramEnsemble.diagram_names, product((0, 1, 2), (0, 1)))
	}
	plot_titles = {
		"ker": "Kernel",
		"cok": "Cokernel",
		"im": "Image",
		"dom": "Domain",
		"cod": "Codomain",
		"rel": "Relative",
	}
	points_legend = {
		name: (True if name == "rel" else False) for name in DiagramEnsemble.diagram_names
	}
	if truncation is None:
		et = []
		for diagram_name in DiagramEnsemble.diagram_names:
			m = max_dim[diagram_name]
			et += [
				dgms.entrance_times[d]
				for _, d in getattr(dgms, diagram_name).paired
				if dgms.dimensions[d] <= m
			]
		truncation = _get_truncation(et)
	fig, axes = plt.subplots(nrows=2, ncols=3, figsize=[3.5 * 3, 3.5 * 2], sharex=True, sharey=True)
	for diagram_name in DiagramEnsemble.diagram_names:
		_plot_diagram(
			getattr(dgms, diagram_name),
			dgms.entrance_times,
			dgms.dimensions,
			truncation,
			ax=axes[plot_pos[diagram_name]],
			max_dim=max_dim[diagram_name],
			dim_shift=dim_shift[diagram_name],
			title=plot_titles[diagram_name],
			points_legend=points_legend[diagram_name],
			tolerance=tolerance,
		)
	legends = dict()
	for ax in fig.axes:
		handles, labels = ax.get_legend_handles_labels()
		for handle, label in zip(handles, labels):
			legends[label] = handle
		ax.tick_params(labelleft=True, labelbottom=True)
	axes[plot_pos["rel"]].get_legend().remove()
	fig.legend(handles=legends.values(), loc="upper center", bbox_to_anchor=(0.5, 0.05), ncol=5)
	return fig, axes


@interpolate_docstring()
def plot_diagram(
	dgms: DiagramEnsemble,
	diagram_name: str,
	truncation: float | None = None,
	max_diagram_dimension: int | None = None,
	ax: Axes | None = None,
	tolerance: float = 0,
) -> Axes:
	"""
	Plot a specific diagram from a 6-pack.

	Args:
		dgms                  : The 6-pack of persistence diagrams.
		diagram_name          : One of ``${str(DiagramEnsemble.diagram_names)}``.
		truncation            : The maximum entrance time for which the diagrams are plotted. A sensible default will be calculated if not provided.
		max_diagram_dimension : The maximum homological dimension for which to plot points. If not provided, all dimensions will be included in the plots.
		ax                    : A matplotlib axes object. If provided then the diagram will be plotted on the given axes.
		tolerance             : Only features with persistence greater than this value will be plotted.
	"""
	if diagram_name not in DiagramEnsemble.diagram_names:
		raise KeyError("Invalid diagram name!")
	titles = {
		"ker": "Kernel",
		"cok": "Cokernel",
		"im": "Image",
		"dom": "Domain",
		"cod": "Codomain",
		"rel": "Relative",
	}
	if max_diagram_dimension is None:
		max_diagram_dimension = max(dgms.dimensions)
	if truncation is None:
		et = [
			dgms.entrance_times[d]
			for _, d in getattr(dgms, diagram_name).paired
			if dgms.dimensions[d] <= max_diagram_dimension
		]
		truncation = _get_truncation(et)
	if ax is None:
		ax1: Axes
		_, ax1 = plt.subplots()
	else:
		ax1 = ax
	_plot_diagram(
		getattr(dgms, diagram_name),
		dgms.entrance_times,
		dgms.dimensions,
		truncation,
		ax=ax1,
		dim_shift=1 if diagram_name == "ker" else 0,
		points_legend=True,
		lines_legend=True,
		title=titles[diagram_name],
		max_dim=max_diagram_dimension,
		tolerance=tolerance,
	)
	return ax1


def _plot_diagram(
	diagram,
	entrance_times,
	dimensions,
	truncation,
	max_dim=2,
	ax=None,
	title=None,
	points_legend=False,
	lines_legend=False,
	dim_shift=0,
	tolerance=0.0,
	**kwargs,
) -> None:
	# Truncate all times
	truncate_level = truncation * 1.04
	inf_level = truncation * 1.08
	entrance_times = [
		et if et <= truncation else truncate_level if et is not np.inf else inf_level
		for et in entrance_times
	]
	all_pts = [
		(
			entrance_times[birth_idx],
			entrance_times[death_idx],
			f"$H_{{{dimensions[birth_idx] - dim_shift}}}$",
		)
		for birth_idx, death_idx in diagram.paired
		if dimensions[birth_idx] - dim_shift <= max_dim
		and entrance_times[death_idx] - entrance_times[birth_idx] > tolerance
	] + [
		(
			entrance_times[birth_idx],
			inf_level,
			f"$H_{{{dimensions[birth_idx] - dim_shift}}}$",
		)
		for birth_idx in diagram.unpaired
		if dimensions[birth_idx] - dim_shift <= max_dim
	]
	df = DataFrame(data=all_pts, columns=["Birth", "Death", "Dimension"])
	df["Dimension"] = df["Dimension"].astype("category")
	ret_ax = sns.scatterplot(
		data=df, x="Birth", y="Death", hue="Dimension", ax=ax, legend=points_legend, **kwargs
	)
	ret_ax.set(xlabel=None)
	ret_ax.set(ylabel=None)
	ret_ax.set(title=title)
	ret_ax.set_aspect("equal")
	if points_legend:
		sns.move_legend(ret_ax, "lower right")
	handle = ax if ax is not None else plt
	handle.plot([0, inf_level], [0, inf_level], "k-", alpha=0.4)
	handle.plot(
		[0, truncate_level], [truncate_level, truncate_level], "k:", alpha=0.4, label="Truncated"
	)
	handle.plot([0, inf_level], [inf_level, inf_level], "k--", alpha=0.4, label="$\\infty$")
	handle.plot([0, 0], [0, inf_level], "k-", alpha=0.4)
	if lines_legend:
		handle.legend()


def draw_filtration(
	K: FilteredComplex,
	points: Annotated[np.ndarray, np.float64],
	time: float,
	include_colours: Collection[int] | None = None,
	ax: Axes | None = None,
) -> Axes:
	"""
	Visualise a filtration at given time, optionally including only certain colours.

	Args:
		K               : A filtered complex.
		points          : The vertices of ``K`` as a :math:`2\\times N` numpy matrix.
		time            : Filtration times for which to draw simplices.
		include_colours : Optional collection of colours to include. If not specified then all colours will be drawn.
		ax              : A matplotlib axes object. If provided then the diagram will be plotted on the given axes.
	"""
	if len(points.shape) != 2:
		raise NotImplementedError

	if include_colours is None:
		include_colours = set(
			[_bitmask_to_colours(vertex.colours)[0] for vertex in K.simplices[0].values()]
		)

	include_colours_bitmask = _colours_to_bitmask(include_colours)

	if ax is None:
		ax1: Axes
		_, ax1 = plt.subplots()
	else:
		ax1 = ax

	plot_colours = np.array(plt.rcParams["axes.prop_cycle"].by_key()["color"])

	# Plot the vertices
	vertex_colours = []
	vertices_to_plot = []
	for idx, vertex in K.simplices[0].items():
		if (
			_colours_are_subset(vertex.colours, include_colours_bitmask)
		) and vertex.filtration_value <= time:
			vertex_colours += _bitmask_to_colours(vertex.colours)
			vertices_to_plot.append(idx)

	vertex_colours = np.array(vertex_colours)
	ax1.scatter(
		points[0, vertices_to_plot],
		points[1, vertices_to_plot],
		c=list(plot_colours[vertex_colours]),
		s=10,
	)

	# Plot the edges
	for idx, simplex in K.simplices[1].items():
		if (
			_colours_are_subset(simplex.colours, include_colours_bitmask)
			and simplex.filtration_value <= time
		):
			if _num_colours_in_bitmask(simplex.colours) == 1:
				colour = plot_colours[_bitmask_to_colours(simplex.colours)[0]]
				alpha = 0.5
			else:
				colour = "black"
				alpha = 0.2
			ax1.plot(
				points[0, simplex.vertices],
				points[1, simplex.vertices],
				c=colour,
				alpha=alpha,
				linewidth=1,
			)

	for idx, simplex in K.simplices[2].items():
		if (
			_colours_are_subset(simplex.colours, include_colours_bitmask)
			and simplex.filtration_value <= time
		):
			if _num_colours_in_bitmask(simplex.colours) == 1:
				colour = plot_colours[_bitmask_to_colours(simplex.colours)[0]]
			else:
				colour = "grey"
			ax1.fill(points[0, simplex.vertices], points[1, simplex.vertices], c=colour, alpha=0.2)

	ax1.set_aspect("equal")
	# ax.set_xlabel('Time = ' + f"{0.0:.4f}")

	return ax1


def animate_filtration(
	K: FilteredComplex,
	points: Annotated[np.ndarray, np.float64],
	filtration_times: Sequence[float],
	animation_length: float,
) -> animation.FuncAnimation:
	"""
	Create animation of 2-skeleton of filtered simplicial complex.

	Args:
		K                : A filtered complex.
		points           : The vertices of ``K`` as a :math:`2\\times N` numpy matrix.
		filtration_times : Sequence of filtration times for which to draw animation frames.
		animation_length : Total length of the animation in seconds.
	"""
	if len(points.shape) != 2:
		raise NotImplementedError

	fig: Figure
	ax: Axes
	fig, ax = plt.subplots()
	plot_colours = np.array(plt.rcParams["axes.prop_cycle"].by_key()["color"])
	vertex_colours = []
	for idx, simplex in K.simplices[0].items():
		i = _bitmask_to_colours(simplex.colours)[0]
		vertex_colours.append(i)

	ax.scatter(points[0, :], points[1, :], c=list(plot_colours[vertex_colours]), s=10)

	lines = dict()
	patches = dict()
	for idx, simplex in K.simplices[1].items():
		if _num_colours_in_bitmask(simplex.colours) == 1:
			i = _bitmask_to_colours(simplex.colours)[0]
			colour = plot_colours[i]
			alpha = 0.5
		else:
			colour = "black"
			alpha = 0.2
		lines[idx] = ax.plot([], [], c=colour, alpha=alpha, linewidth=1)[0]

	for idx, simplex in K.simplices[2].items():
		if _num_colours_in_bitmask(simplex.colours) == 1:
			i = _bitmask_to_colours(simplex.colours)[0]
			colour = plot_colours[i]
		else:
			colour = "grey"
		patches[idx] = ax.fill([], [], c=colour, alpha=0.2)[0]

	ax.set_aspect("equal")
	ax.set_xlabel("Time = " + f"{0.0:.4f}")

	def update(time: float) -> None:
		for idx, simplex in K.simplices[1].items():
			if simplex.filtration_value <= time:
				lines[idx].set_xdata(points[0, simplex.vertices])
				lines[idx].set_ydata(points[1, simplex.vertices])

		for idx, simplex in K.simplices[2].items():
			if simplex.filtration_value <= time:
				patches[idx].set_xy(points[:, simplex.vertices].T)

		ax.set_xlabel(f"Time = {time:.4f}")

	interval = int(np.round(animation_length * 1000 / len(filtration_times)))
	return animation.FuncAnimation(fig=fig, func=update, frames=filtration_times, interval=interval)


def _get_truncation(entrance_times: list[float]) -> float:
	return (
		ub * 2
		if (m := max(entrance_times))
		> 2
		* (
			ub := np.percentile(
				entrance_times,
				[
					95,
				],
			)[0]
		)
		else m
	)



from __future__ import annotations

from collections.abc import Collection, Sequence
from dataclasses import dataclass, field, fields
from typing import Annotated, Callable, ClassVar, overload

import numpy as np
from h5py import Dataset, Group
from phimaker import compute_ensemble

from chalc.chromatic import alpha, delcech, delrips
from chalc.filtration import FilteredComplex

from ._utils import interpolate_docstring

__doc__ = "Routines for computing 6-packs of persistence diagrams."
__all__ = [
	"from_filtration",
	"compute",
	"SimplexPairings",
	"DiagramEnsemble",
	"save_diagrams",
	"load_diagrams",
]

ChromaticMethod = {
	"chromatic alpha": alpha,
	"chromatic delcech": delcech,
	"chromatic delrips": delrips,
}

BdMatType = list[tuple[bool, int, list[int]]]


# bitmask that represents a list of colours
def _colours_to_bitmask(colours: Collection[int]) -> int:
	return sum(2**i for i in colours)


# list of colours represented by a bitmask
def _bitmask_to_colours(b: int) -> list[int]:
	i = 0
	res = []
	while b:
		if b & 1:
			res.append(i)
		i += 1
		b >>= 1
	return res


# number of colours specified by a bitmask
def _num_colours_in_bitmask(b: int) -> int:
	return len(_bitmask_to_colours(b))


# check if the colours specified by bitmask b1 are a subset of those specified by b2
def _colours_are_subset(b1: int, b2: int) -> bool:
	return not (~b2 & b1)


def _get_diagrams(
	K: FilteredComplex,
	check_in_domain: Callable[[int], bool],
	max_dgm_dim: int,
	tolerance: float = 0,
) -> DiagramEnsemble:
	# Build up the matrix
	matrix: BdMatType = []
	entrance_times: list[float] = []
	dimensions: list[int] = []
	for column in K.serialised():
		facet_idxs = column[0]
		dimension = max(0, len(facet_idxs) - 1)
		# Only need k-skeleton for k <= max_dgm_dim + 1
		if dimension > max_dgm_dim + 1:
			break
		dimensions.append(dimension)
		entrance_times.append(column[2])
		matrix.append((check_in_domain(column[3]), dimension, facet_idxs))
	d = compute_ensemble(matrix)
	dgms = DiagramEnsemble(
		SimplexPairings._fromPhimaker(d.ker),
		SimplexPairings._fromPhimaker(d.cok),
		SimplexPairings._fromPhimaker(d.g),
		SimplexPairings._fromPhimaker(d.f),
		SimplexPairings._fromPhimaker(d.im),
		SimplexPairings._fromPhimaker(d.rel),
		entrance_times,
		dimensions,
	)
	return dgms.threshold(tolerance)


@interpolate_docstring()
def from_filtration(
	K: FilteredComplex,
	dom: Collection[int] | int | None = None,
	k: int | None = None,
	max_diagram_dimension: int | None = None,
	tolerance: float = 0,
) -> DiagramEnsemble:
	"""
	Compute the 6-pack of persistence diagrams associated to a filtered simplicial complex with coloured vertices.

	Given a filtered chromatic simplicial complex :math:`K` and a subcomplex :math:`L` of :math:`K`, this function computes the 6-pack of persistence diagram associated with the inclusion map :math:`f : L \\hookrightarrow K`. The subcomplex is specified by
	the colours of its vertices, or by an integer :math:`k` wherein all simplices with :math:`k` or fewer colours are considered part of the subcomplex.

	Args:
		K                     : A filtered chromatic simplicial complex.
		dom                   : Integer or collection of integers describing the colours of the points in the domain (the subcomplex :math:`L`).
		k                     : If not ``None``, then the domain is taken to be the :math:`k`-chromatic subcomplex of :math:`K`, i.e., the subcomplex of simplices having at most :math:`k` colours.
		max_diagram_dimension : Maximum homological dimension for which the persistence diagrams are computed. By default diagrams of all dimensions are computed.
		tolerance             : Retain only points with persistence strictly greater than this value.

	Returns:
		Diagrams corresponding to the following persistence modules (where :math:`H_*` is the persistent homology functor and :math:`f_*` is the induced map on persistent homology):

		#. :math:`H_*(L)` (domain)
		#. :math:`H_*(K)` (codomain)
		#. :math:`\\ker(f_*)` (kernel)
		#. :math:`\\mathrm{coker}(f_*)` (cokernel)
		#. :math:`\\mathrm{im}(f_*)` (image)
		#. :math:`H_*(K, L)` (relative homology)

		Each diagram is represented by sets of paired and unpaired simplices,
		and contain simplices of all dimensions. ``dgms`` also contains the
		entrance times of the simplices and their dimensions.
	"""
	if k is None and dom is None:
		raise RuntimeError("At least one of k or dom must be provided")
	if k is not None and dom is not None:
		raise RuntimeError("Only one of k or dom is allowed")
	if dom is not None:
		if isinstance(dom, int):
			dom = [
				dom,
			]
		colours_bitmask = _colours_to_bitmask(dom)

		def check_in_domain(b):
			return _colours_are_subset(b, colours_bitmask)
	elif k is not None:

		def check_in_domain(b):
			return _num_colours_in_bitmask(b) <= k  # k-chromatic simplex

	# if(K, L) is a pair of simplicial complexes where dim(L) <= dim(K) = d
	# then all of the sixpack diagrams except for the relative homology
	# can be non-zero upto dimension d, and relative can be nontrivial upto dim = d+1
	if max_diagram_dimension is None:
		max_diagram_dimension = K.dimension + 1
	else:
		max_diagram_dimension = min(max_diagram_dimension, K.dimension + 1)
	return _get_diagrams(K, check_in_domain, max_diagram_dimension, tolerance)


@interpolate_docstring()
def compute(
	x: Annotated[np.ndarray, np.float64],
	colours: Sequence[int],
	dom: Collection[int] | int | None = None,
	k: int | None = None,
	method: str = "chromatic alpha",
	max_diagram_dimension: int | None = None,
	tolerance: float = 0,
) -> DiagramEnsemble:
	"""
	Compute the 6-pack of persistence diagrams of a coloured point-cloud.

	This function constructs a filtered simplicial complex :math:`K` from the point cloud, and computes the 6-pack of persistence diagrams associated with the inclusion :math:`f : L \\hookrightarrow K` where :math:`L` is some filtered subcomplex of :math:`K`.

	Args:
		x                     : Numpy matrix whose columns are points.
		colours               : Sequence of integers describing the colours of the points.
		dom                   : Integer or collection of integers describing the colours of the points in the domain (the subcomplex :math:`L`).
		k                     : If not ``None``, then the domain is taken to be the :math:`k`-chromatic subcomplex of :math:`K`, i.e., the subcomplex of simplices having at most :math:`k` colours.
		method                : Filtration used to construct the chromatic complex. Must be one of ``${str(list(ChromaticMethod.keys()))}``.
		max_diagram_dimension : Maximum homological dimension for which the persistence diagrams are computed. By default diagrams of all dimensions are computed.
		tolerance             : Retain only points with persistence strictly greater than this value.

	Returns :
		Diagrams corresponding to the following persistence modules (where :math:`H_*` is the persistent homology functor and :math:`f_*` is the induced map on persistent homology):

		#. :math:`H_*(L)` (domain)
		#. :math:`H_*(K)` (codomain)
		#. :math:`\\ker(f_*)` (kernel)
		#. :math:`\\mathrm{coker}(f_*)` (cokernel)
		#. :math:`\\mathrm{im}(f_*)` (image)
		#. :math:`H_*(K, L)` (relative homology)

		Each diagram is represented by lists of paired and unpaired simplices,
		and contains simplices of all dimensions. ``dgms`` also contains the
		entrance times of the simplices and their dimensions.
	"""
	if k is None and dom is None:
		raise RuntimeError("At least one of k or dom must be provided")
	if k is not None and dom is not None:
		raise RuntimeError("Only one of k or dom is allowed")
	if dom is not None:
		# new colours: 1 -> domain, 0 -> codomain
		if isinstance(dom, int):
			dom = [
				dom,
			]
		new_colours = list(np.isin(colours, list(dom)).astype(np.int64))

		def check_in_domain(b: int):
			return b == 2
	elif k is not None:

		def check_in_domain(b: int):
			return _num_colours_in_bitmask(b) <= k  # k-chromatic simplex

		new_colours = list(colours)
	# can have non-zero homology upto dimension d for everything except rel
	# rel can have non-zero homology in dimension d+1
	if max_diagram_dimension is None:
		max_diagram_dimension = x.shape[0]
	else:
		max_diagram_dimension = min(max_diagram_dimension, x.shape[0])
	# Compute chromatic complex
	if method in ChromaticMethod.keys():
		K, _ = ChromaticMethod[method](x, new_colours)
	else:
		raise RuntimeError(f"method must be one of {ChromaticMethod.keys()}")
	return _get_diagrams(K, check_in_domain, max_diagram_dimension, tolerance)


@dataclass
class SimplexPairings:
	"""
	Persistence diagram object, represented by a
	list of simplex pairings and a list of unpaired simplices.
	"""

	#: Set of tuples of indices of paired simplices.
	paired: list[tuple[int, int]] = field(default_factory=list)
	#: Set of indices of unpaired simplices.
	unpaired: list[int] = field(default_factory=list)

	@classmethod
	def _fromPhimaker(cls, obj) -> SimplexPairings:
		return cls(list(obj.paired), list(obj.unpaired))

	def __str__(self):
		return f"Paired: {self.paired}\nUnpaired: {self.unpaired}"

	@classmethod
	def _from_matrices(
		cls, p: Annotated[np.ndarray, np.int64], u: Annotated[np.ndarray, np.int64]
	) -> SimplexPairings:
		if p.shape[1] != 2:
			raise ValueError(f"p must be a (m, 2) matrix, but received a matrix of size {p.shape}")
		paired: list[tuple[int, int]] = [(p[i, 0], p[i, 1]) for i in range(p.shape[0])]
		unpaired = list(u)
		return cls(paired, unpaired)

	def paired_as_matrix(self) -> Annotated[np.ndarray, np.int64]:
		return np.concatenate(list(self.paired)).reshape((-1, 2))

	def __eq__(self, other) -> bool:
		if isinstance(other, SimplexPairings):
			return set(self.paired) == set(other.paired) and set(self.unpaired) == set(
				other.unpaired
			)
		else:
			return NotImplemented


@dataclass
class DiagramEnsemble:
	"""
	6-pack of persistence diagrams.
	"""

	# List of matrices
	diagram_names: ClassVar[tuple[str, ...]] = ("ker", "cok", "dom", "cod", "im", "rel")
	#: Kernel
	ker: SimplexPairings = field(default_factory=lambda: SimplexPairings())
	#: Cokernel
	cok: SimplexPairings = field(default_factory=lambda: SimplexPairings())
	#: Domain
	dom: SimplexPairings = field(default_factory=lambda: SimplexPairings())
	#: Codomain
	cod: SimplexPairings = field(default_factory=lambda: SimplexPairings())
	#: Image
	im: SimplexPairings = field(default_factory=lambda: SimplexPairings())
	#: Relative
	rel: SimplexPairings = field(default_factory=lambda: SimplexPairings())
	#: Entrance times of the simplices
	entrance_times: list[float] = field(default_factory=lambda: [])
	#: Dimensions of the simplices
	dimensions: list[int] = field(default_factory=lambda: [])

	def __eq__(self, other) -> bool:
		if isinstance(other, DiagramEnsemble):
			return all(
				getattr(self, f.name) == getattr(other, f.name) for f in fields(DiagramEnsemble)
			)
		else:
			return NotImplemented

	def threshold(self, tolerance: float) -> DiagramEnsemble:
		"""
		Discard all features with persistence ``<=tolerance``.
		"""
		for name in DiagramEnsemble.diagram_names:
			pairings = [
				(b, d)
				for b, d in getattr(self, name).paired
				if self.entrance_times[d] - self.entrance_times[b] > tolerance
			]
			setattr(getattr(self, name), "paired", pairings)
		return self

	@overload
	def get(self, diagram_name: str, dim: int) -> Annotated[np.ndarray, np.float64]: ...

	@overload
	def get(
		self, diagram_name: str, dim: list[int] | None = None
	) -> list[Annotated[np.ndarray, np.float64]]: ...

	@interpolate_docstring({"diagram_names": diagram_names})
	def get(self, diagram_name, dim=None):
		"""
		Get a specific diagram as a matrix of birth and death times.

		Args:
			diagram_name : One of ``${str(diagram_names)}``.
			dim          : Dimension(s) of the diagram desired. If a list is provided then a list of matrices is returned, with the order of matrices respecting the order of entries of `dim`. If `dim` is not provided then the returned matrix will contain persistent features from all homological dimensions from zero to ``max(self.dimensions)``.

		Returns:
			An :math:`m \\times 2` matrix whose rows are a pair of birth and death times, or a list of such matrices.
		"""
		if dim is None:
			dim = list(range(max(set(self.dimensions))))
		if isinstance(dim, int):
			# a p-dimensional homology class is captured by a pairing of (p+1) simplices for kernels
			if diagram_name == "ker":
				dim += 1
			diagram = getattr(self, diagram_name)
			entrance_times = np.array(self.entrance_times)
			paired_list = np.array(
				[pair for pair in diagram.paired if self.dimensions[pair[0]] == dim], dtype=int
			)
			unpaired_list = np.array(
				[sigma for sigma in diagram.unpaired if self.dimensions[sigma] == dim], dtype=int
			)
			paired_matrix = entrance_times[paired_list.flatten()].reshape((-1, 2))
			unpaired_matrix = entrance_times[unpaired_list][:, np.newaxis]
			inf_array = np.array([np.inf] * len(unpaired_list))[:, np.newaxis]
			unpaired_matrix = np.concatenate((unpaired_matrix, inf_array), axis=1)
			return np.concatenate((paired_matrix, unpaired_matrix), axis=0)
		elif isinstance(dim, list):
			return [self.get(diagram_name, d) for d in dim]


def save_diagrams(dgms: DiagramEnsemble, file: Group) -> None:
	"""
	Save a 6-pack of persistence diagrams to a HDF5 file or group.

	Args:
		dgms : 6-pack of diagrams to save to file/group.
		file : A h5py file or group.
	"""
	file["entrance_times"] = np.array(dgms.entrance_times, dtype=np.float64)
	file["dimensions"] = np.array(dgms.dimensions, dtype=np.int64)
	for diagram_name in DiagramEnsemble.diagram_names:
		grp = file.require_group(diagram_name)
		grp["unpaired"] = np.array(list(getattr(dgms, diagram_name).unpaired), dtype=np.int64)
		grp["paired"] = getattr(dgms, diagram_name).paired_as_matrix()


def load_diagrams(file: Group) -> DiagramEnsemble:
	"""
	Load a 6-pack of persistence diagrams from a HDF5 file or group.
	"""
	dgms = DiagramEnsemble()
	if any(f.name not in file for f in fields(DiagramEnsemble)):
		raise RuntimeError("Invalid file: missing fields")

	dgms.entrance_times = (
		list(entrance_times)
		if isinstance((entrance_times := file["entrance_times"]), Dataset)
		and entrance_times.dtype == np.float64
		else []
	)
	dgms.dimensions = (
		list(dimensions)
		if isinstance((dimensions := file["dimensions"]), Dataset) and dimensions.dtype == np.int64
		else []
	)
	for diagram_name in DiagramEnsemble.diagram_names:
		grp = file[diagram_name]
		if isinstance(grp, Group) and all(
			name in list(grp.keys()) for name in ["paired", "unpaired"]
		):
			paired = grp["paired"]
			unpaired = grp["unpaired"]
			if (
				isinstance(paired, Dataset)
				and paired.dtype == np.int64
				and isinstance(unpaired, Dataset)
				and unpaired.dtype == np.int64
			):
				setattr(
					dgms, diagram_name, SimplexPairings._from_matrices(paired[...], unpaired[...])
				)
			else:
				raise RuntimeError("Invalid file: missing fields")
	return dgms



__version__ = '0.1.0'
from sphinx.ext.autosummary.generate import(
    AutosummaryRenderer, 
    ModuleScanner, 
    _get_members, 
    _get_module_attrs,
    _get_modules)
from sphinx.ext.autosummary import get_documenter
from typing import Any

# autosummary does not correctly get parse class properties
# this fixes that problem
def generate_autosummary_content(name: str, obj: Any, parent: Any,
                                 template: AutosummaryRenderer, template_name: str,
                                 imported_members: bool, app: Any,
                                 recursive: bool, context: dict,
                                 modname: str | None = None,
                                 qualname: str | None = None) -> str:
    doc = get_documenter(app, obj, parent)

    ns: dict[str, Any] = {}
    ns.update(context)

    if doc.objtype == 'module':
        scanner = ModuleScanner(app, obj)
        ns['members'] = scanner.scan(imported_members)

        respect_module_all = not app.config.autosummary_ignore_module_all
        imported_members = imported_members or ('__all__' in dir(obj) and respect_module_all)

        ns['functions'], ns['all_functions'] = \
            _get_members(doc, app, obj, {'function'}, imported=imported_members)
        ns['classes'], ns['all_classes'] = \
            _get_members(doc, app, obj, {'class'}, imported=imported_members)
        ns['exceptions'], ns['all_exceptions'] = \
            _get_members(doc, app, obj, {'exception'}, imported=imported_members)
        ns['attributes'], ns['all_attributes'] = \
            _get_module_attrs(name, ns['members'])
        ispackage = hasattr(obj, '__path__')
        if ispackage and recursive:
            # Use members that are not modules as skip list, because it would then mean
            # that module was overwritten in the package namespace
            skip = (
                ns["all_functions"]
                + ns["all_classes"]
                + ns["all_exceptions"]
                + ns["all_attributes"]
            )

            # If respect_module_all and module has a __all__ attribute, first get
            # modules that were explicitly imported. Next, find the rest with the
            # get_modules method, but only put in "public" modules that are in the
            # __all__ list
            #
            # Otherwise, use get_modules method normally
            if respect_module_all and '__all__' in dir(obj):
                imported_modules, all_imported_modules = \
                    _get_members(doc, app, obj, {'module'}, imported=True)
                skip += all_imported_modules
                imported_modules = [name + '.' + modname for modname in imported_modules]
                all_imported_modules = \
                    [name + '.' + modname for modname in all_imported_modules]
                public_members = getall(obj)
            else:
                imported_modules, all_imported_modules = [], []
                public_members = None

            modules, all_modules = _get_modules(obj, skip=skip, name=name,
                                                public_members=public_members)
            ns['modules'] = imported_modules + modules
            ns["all_modules"] = all_imported_modules + all_modules
    elif doc.objtype == 'class':
        ns['members'] = dir(obj)
        ns['inherited_members'] = \
            set(dir(obj)) - set(obj.__dict__.keys())
        ns['methods'], ns['all_methods'] = \
            _get_members(doc, app, obj, {'method'}, include_public={'__init__'})
        ns['attributes'], ns['all_attributes'] = \
            _get_members(doc, app, obj, {'attribute'})
        ns['properties'], ns['all_properties'] = \
            _get_members(doc, app, obj, {'property'})
        ns['classes'], ns['all_classes'] = \
            _get_members(doc, app, obj, {'class'})

    if modname is None or qualname is None:
        modname, qualname = _split_full_qualified_name(name)

    if doc.objtype in ('method', 'attribute', 'property'):
        ns['class'] = qualname.rsplit(".", 1)[0]

    if doc.objtype in ('class',):
        shortname = qualname
    else:
        shortname = qualname.rsplit(".", 1)[-1]

    ns['fullname'] = name
    ns['module'] = modname
    ns['objname'] = qualname
    ns['name'] = shortname

    ns['objtype'] = doc.objtype
    ns['underline'] = len(name) * '='

    if template_name:
        return template.render(template_name, ns)
    else:
        return template.render(doc.objtype, ns)


def autosummary_modify_content(func):
    import sphinx.ext.autosummary

    sphinx.ext.autosummary.generate.generate_autosummary_content = func

def setup(app):
    app.setup_extension("sphinx.ext.autosummary")
    autosummary_modify_content(generate_autosummary_content)

    return {"version": __version__, "parallel_read_safe": True}


import itertools
import math

import chalc as ch
import numpy as np


class Test_chromatic:
	# equilateral triangle centred at origin
	# plus a point at the origin
	points = np.array(
		[
			[0, 0],
			[1, 0],
			[np.cos(2 / 3 * np.pi), np.sin(2 / 3 * np.pi)],
			[np.cos(4 / 3 * np.pi), np.sin(4 / 3 * np.pi)],
		]
	).T
	colours = [1, 0, 0, 0]
	random_seed = 44

	def test_delaunay(self):
		K = ch.chromatic.delaunay(self.points, self.colours)
		assert_standard_simplex(3, K, filtered=False, chromatic=True)
		assert K.simplices[0][0].colours == 2
		assert all(K.simplices[0][i].colours == 1 for i in range(1, 4))
		assert all(K.simplices[1][i].colours == 3 for i in range(3))
		assert all(K.simplices[1][i].colours == 1 for i in range(3, 6))
		assert all(K.simplices[2][i].colours == 3 for i in range(3))
		assert K.simplices[3][0].colours == 3

	def test_alpha(self):
		K, numerical_errors = ch.chromatic.alpha(self.points, self.colours)
		assert not numerical_errors
		assert K.is_filtration()
		assert all(K.simplices[0][i].filtration_value == 0 for i in range(4))
		assert all(math.isclose(K.simplices[1][i].filtration_value, 0.5) for i in range(3))
		assert all(
			math.isclose(K.simplices[1][i].filtration_value, np.sqrt(3) / 2) for i in range(3, 6)
		)
		assert all(
			math.isclose(K.simplices[2][i].filtration_value, np.sqrt(3) / 2) for i in range(3)
		)
		assert math.isclose(K.simplices[2][3].filtration_value, 1)
		assert math.isclose(K.simplices[3][0].filtration_value, 1)

	def test_delcech(self):
		K, numerical_errors = ch.chromatic.delcech(self.points, self.colours)
		assert not numerical_errors
		assert K.is_filtration()
		assert all(K.simplices[0][i].filtration_value == 0 for i in range(4))
		assert all(math.isclose(K.simplices[1][i].filtration_value, 0.5) for i in range(3))
		assert all(
			math.isclose(K.simplices[1][i].filtration_value, np.sqrt(3) / 2) for i in range(3, 6)
		)
		assert all(
			math.isclose(K.simplices[2][i].filtration_value, np.sqrt(3) / 2) for i in range(3)
		)
		assert math.isclose(K.simplices[2][3].filtration_value, 1)
		assert math.isclose(K.simplices[3][0].filtration_value, 1)

	def test_delrips(self):
		K, numerical_errors = ch.chromatic.delrips(self.points, self.colours)
		assert not numerical_errors
		assert K.is_filtration()
		assert all(K.simplices[0][i].filtration_value == 0 for i in range(4))
		assert all(math.isclose(K.simplices[1][i].filtration_value, 0.5) for i in range(3))
		assert all(
			math.isclose(K.simplices[1][i].filtration_value, np.sqrt(3) / 2) for i in range(3, 6)
		)
		assert all(
			math.isclose(K.simplices[2][i].filtration_value, np.sqrt(3) / 2) for i in range(4)
		)
		assert math.isclose(K.simplices[3][0].filtration_value, np.sqrt(3) / 2)

	def test_delcech_is_filtration(self):
		rng = np.random.default_rng(self.random_seed)
		dims = [1, 2, 3]
		num_colours = [1, 2, 3]
		for d in dims:
			for s in num_colours:
				points = rng.uniform(size=(d, 200))
				colours = rng.integers(0, s, size=200)
				K, numerical_errors = ch.chromatic.delcech(points, colours)
				assert not numerical_errors
				assert(K.is_filtration())

	def test_alpha_is_filtration(self):
		rng = np.random.default_rng(self.random_seed)
		dims = [1, 2, 3]
		num_colours = [1, 2, 3]
		for d in dims:
			for s in num_colours:
				points = rng.uniform(size=(d, 200))
				colours = rng.integers(0, s, size=200)
				K, numerical_errors = ch.chromatic.alpha(points, colours)
				assert not numerical_errors
				assert(K.is_filtration())

	def test_delrips_is_filtration(self):
		rng = np.random.default_rng(self.random_seed)
		dims = [1, 2, 3]
		num_colours = [1, 2, 3]
		for d in dims:
			for s in num_colours:
				points = rng.uniform(size=(d, 200))
				colours = rng.integers(0, s, size=200)
				K, numerical_errors = ch.chromatic.delrips(points, colours)
				assert not numerical_errors
				assert(K.is_filtration())

# class Test_sixpack:
# 	random_seed = 44
# 	def test_k_chromatic(self):
# 		rng = np.random.default_rng(self.random_seed)
# 		points = rng.uniform(size=(2, 1000))
# 		colours = rng.integers(0, 3, size=1000)
# 		dgms = ch.sixpack.compute(points, colours, method='chromatic delcech')


class Test_filtration:
	def test_standard_simplex(self):
		n = 3
		K = ch.filtration.standard_simplex(n)
		assert_standard_simplex(n, K, filtered=False, chromatic=False)

	def test_clique_complex(self):
		n = 5
		k = 3
		K = ch.filtration.clique_complex(n, k)
		for i in range(k + 1):
			num_simplices = math.comb(n, i + 1)
			assert list(K.simplices[i].keys()) == list(range(num_simplices))
			gen_vertices = itertools.combinations(range(n), i + 1)
			for j in range(num_simplices):
				assert K.simplices[i][j].dimension == i
				assert K.simplices[i][j].vertices == list(next(gen_vertices))
				assert K.simplices[i][j].colours == 1
				assert K.simplices[i][j].filtration_value == 0
				assert K.simplices[i][j].label == j


def assert_standard_simplex(
	n: int, K: ch.filtration.FilteredComplex, filtered: bool, chromatic: bool
):
	assert K.dimension == n
	assert len(K.simplices) == n + 1
	for i in range(n + 1):
		num_simplices = math.comb(n + 1, i + 1)
		assert list(K.simplices[i].keys()) == list(range(num_simplices))
		gen_vertices = itertools.combinations(range(n + 1), i + 1)
		for k in range(num_simplices):
			assert K.simplices[i][k].dimension == i
			assert K.simplices[i][k].vertices == list(next(gen_vertices))
			assert K.simplices[i][k].label == k
			if not chromatic:
				assert K.simplices[i][k].colours == 1
			if not filtered:
				assert K.simplices[i][k].filtration_value == 0



from __future__ import annotations

import platform

from . import filtration
if platform.system() == 'Windows':
    # need to load up the GMP and MPFR DLLs
    import os as _os
    from importlib.metadata import files as _files
    _bindir = [x 
        for x in _files('chalc') 
        if len(x.parent.parts) > 0 
        and x.parent.parts[-1] == 'bin'][0].locate().parent
    with _os.add_dll_directory(_bindir):
        from . import chromatic
else:
    from . import chromatic

from . import sixpack
from . import plotting

__all__ = ['chromatic', 'filtration', 'sixpack', 'plotting']

__doc__ = "Chalc is a package to compute the persistent homology of chromatic complexes of points clouds in Euclidean space."
